@@ -264,6 +264,7 @@ static void run_expire(void)
 
 static sig_atomic_t shutdown = 0;
 static sig_atomic_t prune = 0;
+static sig_atomic_t reinit = 0;
 static sig_atomic_t expire_now = 0;
 
 static void sig_shutdown(int sig)
@@ -281,11 +282,17 @@ static void sig_expire(int sig)
   expire_now = 1;
 }
 
+static void sig_reinit(int sig)
+{
+  prune = 1;
+  reinit = 1;
+}
+
 static int get_pkt(int fd, struct autofs_packet_missing *pkt)
 {
   char *buf = (char *) pkt;
   int bytes = sizeof(*pkt);
-  
+
   do {
     int i = read(fd, buf, bytes);
     if (!i)
@@ -301,7 +308,7 @@ static int get_pkt(int fd, struct autofs_packet_missing *pkt)
 	  expire_now = 0;
 	  alarm(ap.exp_runfreq);
 	}
-	if ( !shutdown )
+	if ( !shutdown && !reinit )
 	  continue;
       }
       break;
@@ -336,7 +343,7 @@ static int handle_packet(void)
   struct stat st;
   sigset_t allsignals, oldsig, chldsig;
   pid_t f;
-  
+
   if (get_pkt(ap.pipefd, &pkt))
     return -1;
   
@@ -378,29 +385,16 @@ static int handle_packet(void)
       close(ap.pipefd);
       close(ap.ioctlfd);
       
-      /* Wait for go-ahead from main process.  Fortunately signals sent
-	 by a process to itself are synchronous events per POSIX.1 */
-      kill(getpid(), SIGSTOP);
       i = ap.lookup->lookup_mount(ap.path,pkt.name,pkt.len,ap.lookup->context);
       _exit(i ? 1 : 0);
     } else {
-      int status;
-
-      /* Leave SIGCHLD blocked so the waitpid() below works correctly  */
-      sigaddset(&oldsig, SIGCHLD);
-      sigprocmask(SIG_SETMASK, &oldsig, NULL);
+      /* Important: set up data structures while signals still blocked */
       mt->pid = f;
       mt->wait_queue_token = pkt.wait_queue_token;
       mt->next = ap.mounts;
       ap.mounts = mt;
-      if ( waitpid(f, &status, WUNTRACED) != f ||
-	   !WIFSTOPPED(status) ) {
-	sigprocmask(SIG_UNBLOCK, &chldsig, NULL);
-	send_fail(pkt.wait_queue_token);
-	return 1;
-      }
-      sigprocmask(SIG_UNBLOCK, &chldsig, NULL);
-      kill(f, SIGCONT); /* Go for it */
+
+      sigprocmask(SIG_SETMASK, &oldsig, NULL);
     }
   } else {
     /* Already there (can happen if a process connects to a
@@ -557,6 +551,7 @@ int main(int argc, char *argv[])
   
   program = argv[0];
   
+  memset(&ap, 0, sizeof ap);	/* Initialize ap so we can test for null */
   ap.exp_timeout = DEFAULT_TIMEOUT;
   
   opterr = 0;
@@ -596,7 +591,6 @@ int main(int argc, char *argv[])
     exit(1);
   }
   
-  memset(&ap, 0, sizeof ap);	/* Initialize ap so we can test for null */
   become_daemon();
   
   path    = argv[0];
@@ -626,62 +620,82 @@ int main(int argc, char *argv[])
     cleanup_exit(1);
   }	
   
-  sa.sa_handler = SIG_IGN;
-  sigemptyset(&sa.sa_mask);
-  sa.sa_flags = 0;
-#ifdef SIGUNUSED
-  sigaction(SIGUNUSED, &sa, NULL);
-#endif
-#ifdef SIGLOST
-  sigaction(SIGLOST, &sa, NULL);
-#endif
-  sigaction(SIGVTALRM, &sa, NULL);
-  sigaction(SIGPROF, &sa, NULL);
-  sigaction(SIGPIPE, &sa, NULL);
-  
+  /* The following signals cause a shutdown event to occur */
   sa.sa_handler = sig_shutdown;
   sigemptyset(&sa.sa_mask);
   sigaddset(&sa.sa_mask, SIGCHLD);
+  sigaddset(&sa.sa_mask, SIGALRM);
   sa.sa_flags = 0;
-  sigaction(SIGINT, &sa, NULL);
-  sigaction(SIGQUIT, &sa, NULL);
-  sigaction(SIGILL, &sa, NULL);
-  sigaction(SIGTRAP, &sa, NULL);
-  sigaction(SIGABRT, &sa, NULL);
-  sigaction(SIGIOT, &sa, NULL);
-  sigaction(SIGBUS, &sa, NULL);
-  sigaction(SIGFPE, &sa, NULL);
-  sigaction(SIGSEGV, &sa, NULL);
-  sigaction(SIGTERM, &sa, NULL);
   sigaction(SIGIO, &sa, NULL);
+  sigaction(SIGTERM, &sa, NULL);
+  /* USR2 is special, but is handled by the same handler */
+  sigaction(SIGUSR2, &sa, NULL);
+
+  /* The following signals cause a shutdown event to occur, but if we get
+     more than one, permit the signal to proceed so we don't loop.  This
+     is basically the complete list of "this shouldn't happen" signals. */
+  sa.sa_flags = SA_ONESHOT;
   sigaction(SIGXCPU, &sa, NULL);
   sigaction(SIGXFSZ, &sa, NULL);
-#ifdef SIGSYS
-  sigaction(SIGSYS, &sa, NULL);
-#endif
+  sigaction(SIGINT, &sa, NULL);
 #ifdef SIGPWR
   sigaction(SIGPWR, &sa, NULL);
 #endif
+#ifndef DEBUG
+  /* When debugging, these signals should be in the default state; when
+     in production, we want to at least attempt to catch them and shut down. */
+  sigaction(SIGILL, &sa, NULL);
+  sigaction(SIGQUIT, &sa, NULL);
+  sigaction(SIGABRT, &sa, NULL);
+  sigaction(SIGTRAP, &sa, NULL);
+  sigaction(SIGFPE, &sa, NULL);
+  sigaction(SIGSEGV, &sa, NULL);
+  sigaction(SIGBUS, &sa, NULL);
+  sigaction(SIGPROF, &sa, NULL);
+  sigaction(SIGPIPE, &sa, NULL);
+#ifdef SIGSYS
+  sigaction(SIGSYS, &sa, NULL);
+#endif
 #ifdef SIGSTKFLT
   sigaction(SIGSTKFLT, &sa, NULL);
 #endif
-  sigaction(SIGUSR2, &sa, NULL);
-
+#ifdef SIGLOST
+  sigaction(SIGLOST, &sa, NULL);
+#endif
+#ifdef SIGEMT
+  sigaction(SIGEMT, &sa, NULL);
+#endif
+#endif /* DEBUG */
+  
+  /* The SIGCHLD handler is the bottom handler of the mount request */
   sa.sa_handler = sig_child;
   sigemptyset(&sa.sa_mask);
-  sa.sa_flags = 0;
+  sigaddset(&sa.sa_mask, SIGALRM);
+  sa.sa_flags = SA_NOCLDSTOP;	/* Don't need info about stopped children */
   sigaction(SIGCHLD, &sa, NULL);
   
-  sa.sa_handler = sig_prune;
-  sigemptyset(&sa.sa_mask);
-  sa.sa_flags = 0;
-  sigaction(SIGUSR1, &sa, NULL);
-  sigaction(SIGHUP, &sa, NULL);
-  
+  /* The SIGALRM handler controls expiration of entries. */
   sa.sa_handler = sig_expire;
   sigemptyset(&sa.sa_mask);
   sa.sa_flags = 0;
   sigaction(SIGALRM, &sa, NULL);
+
+  /* SIGUSR1 causes a prune event */
+  sa.sa_handler = sig_prune;
+  sigaction(SIGUSR1, &sa, NULL);
+
+  /* SIGHUP causes complete reinitialization */
+  sa.sa_handler = sig_reinit;
+  sigaction(SIGHUP, &sa, NULL);
+  
+  /* The following signals shouldn't occur, and are ignored */
+  sa.sa_handler = SIG_IGN;
+  sigaction(SIGVTALRM, &sa, NULL);
+  sigaction(SIGURG, &sa, NULL);
+  sigaction(SIGWINCH, &sa, NULL);
+#ifdef SIGUNUSED
+  sigaction(SIGUNUSED, &sa, NULL);
+#endif
   
   /* If this ioctl() doesn't work, it is kernel version 2 */
   if ( ioctl(ap.ioctlfd, AUTOFS_IOC_PROTOVER, &kproto_version) ) {
@@ -713,12 +727,19 @@ int main(int argc, char *argv[])
       break;
     if (shutdown == SIGUSR2 && umount_autofs(0))
       shutdown = 0;
+    if (reinit) {
+      reinit = 0;
+      close_lookup(ap.lookup);
+      if ( !(ap.lookup = open_lookup(map, "", mapfmt, mapargc, mapargv)) ) {
+	syslog(LOG_CRIT, "reinitialization failed, shutting down");
+	umount_autofs(1);
+	cleanup_exit(1);
+      }
+    }
   }
   
   syslog(LOG_INFO, "shutting down, path = %s", path);
-  
-  close_lookup(ap.lookup);
-  umount_autofs(1);
 
+  umount_autofs(1);
   cleanup_exit(0);
 }
