@@ -174,7 +174,7 @@ int lookup_mount(char *root, char *name, int name_len, void *context)
     syslog(LOG_ERR, MODPREFIX "waitpid: %m");
     return 1;
   }
-  if ( !WIFEXITED(status) || WEXITSTATUS(status) != 0 ) {
+  if ( mapp == mapent || !WIFEXITED(status) || WEXITSTATUS(status) != 0 ) {
     syslog(LOG_NOTICE, MODPREFIX "lookup for %s failed", name);
     return 1;
   }
