@@ -360,8 +360,10 @@ int parse_init(int argc, char **argv, void **context)
       kill_context(ctxt);
       return 1;
     }
-  else
+  else {
+    init_ctr++;
     return 0;
+  }
 }
 
 int parse_mount(char *root, char *name, int name_len, char *mapent, void *context)
@@ -485,14 +487,13 @@ int parse_mount(char *root, char *name, int name_len, char *mapent, void *contex
 
 int parse_done(void *context)
 {
-  int rv;
-
-  kill_context((struct parse_context *) context);
+  int rv = 0;
+  struct parse_context *ctxt = (struct parse_context *)context;
 
   if ( --init_ctr == 0 ) {
-    rv = mount_nfs->mount_done(mount_nfs->context);
+    rv = close_mount(mount_nfs);
     mount_nfs = NULL;
-    return rv;
-  } else
-    return 0;
+  }
+  kill_context(ctxt);
+  return rv;
 }
